==> mdcached-1.0.3.tgz: mdcached-1.0.3/mdcached/ServerNetworkThread.cpp <==
/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerNetworkThread.cpp 347 2011-09-20 20:58:41Z ivoras $
 */

#include <sys/param.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <assert.h>
#include <stdlib.h>
#include <pthread.h>
#include <err.h>

#include <list>
#include <iterator>

#include "logger.h"
#include "debug.h"
#include "misc.h"
#include "minilibevent.h"
#include "server.h"
#include "Thread.h"
#include "ServerWorkerThread.h"
#include "ServerNetworkThread.h"
#include "ServerClientConnection.h"


extern unsigned int n_worker_threads;
extern ServerWorkerThread *worker_threads;
extern bool strongly_bind_flag;

void shutdown_connection(ServerClientConnection *conn);

ServerNetworkThread::ServerNetworkThread() : Thread()
{
	CRIT_ASSERT(pthread_mutex_init(&evq_mtx, NULL) == 0);
	CRIT_ASSERT(pthread_cond_init(&evq_cv, NULL) == 0);
	CRIT_ASSERT(ev_valid_queue(evq = ev_create_queue()));
	rr_worker_thread = 0;
}

ServerNetworkThread::~ServerNetworkThread()
{
	pthread_cond_destroy(&evq_cv);
	pthread_mutex_destroy(&evq_mtx);
}

#ifndef MAX_NETWORK_EVENTS
#define MAX_NETWORK_EVENTS 40
#endif

/**
 * Network thread - gathers IO from connections and dispatches it to worker
 * threads, in a bursty way: tries to gather as much as MAX_NETWORK_EVENTS
 * events if possible while holding locks to avoid IPC overhead.
 */
void
ServerNetworkThread::run()
{
	ServerWorkerThread *wthr;
	ev_list evs;
	ev_result evr;
	struct timespec ts_wait;
	ServerClientConnection *conn;

	mc_log(LOG_INFO, "Network thread %p starting", this);

	evs = ev_allocate_list(MAX_NETWORK_EVENTS);
	if (evs == NULL)
		mc_log(LOG_CRIT, "Cannot allocate event list in network thread %p", this);

	ts_wait.tv_sec = 0;
	ts_wait.tv_nsec = 100000000; /* 0.1s, 100 ms */
	DEBUG_STATUS_POINT(st, '0');

	while (!stop_requested) {
		DEBUG_STATUS_POINT(st, 'e');
		evr = ev_get_new_events(evq, evs, MAX_NETWORK_EVENTS, &ts_wait);
		/* Process new events in bulk, adding them to worker threads (if applicable) */
		DEBUG_STATUS_POINT(st, 'E');
		for (int i = 0; i < evr; i++) {
			conn = (ServerClientConnection*) evs[i].udata;
			assert(conn != NULL);
			assert(conn->type == CONN_TYPE_CLIENT_UNIX || conn->type == CONN_TYPE_CLIENT_TCP);

			if (evs[i].filter == EV_READ) {
				/* Handle incoming data */
				DEBUG_STATUS_POINT(st, 'r');
				if (conn->process_read_event(this) == EVPROC_ERROR) {
					ev_remove_fd(evq, conn->socket, FALSE);
					shutdown_connection(conn);
					continue;
				}
			} else if (evs[i].filter == EV_WRITE && conn->outputDataReady()) {
				/* Handle outgoing data */
				DEBUG_STATUS_POINT(st, 'w');
				if (conn->process_write_event() == EVPROC_ERROR) {
					ev_remove_fd(evq, conn->socket, FALSE);
					shutdown_connection(conn);
					continue;
				}
			}
		}

		if (n_worker_threads > 0) {
			wthr = &worker_threads[rr_worker_thread];
			if (wthr->getQueueLocker() == this) {
				wthr->setQueueLocker(NULL);
				wthr->queueUnlock();
				wthr->signalQueueReady();
				if (!strongly_bind_flag) /* If not "strongly bound", rotate to another thread */
					rr_worker_thread = (rr_worker_thread + 1) % n_worker_threads;
			}
		}
	}
	mc_log(LOG_INFO, "Network thread %p exiting", this);
	free(evs);
	DEBUG_STATUS_POINT(st, 'X');
}

==> mdcached-1.0.4.tgz: mdcached-1.0.4/mdcached/ServerNetworkThread.cpp <==
/*-
 * This file is a part of the MDCacheD server
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: ServerNetworkThread.cpp 449 2012-03-20 14:39:42Z ivoras $
 */

#include <sys/param.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <pthread.h>
#include <signal.h>
#include <syslog.h>
#include <assert.h>
#include <stdlib.h>
#include <pthread.h>
#include <err.h>

#include <list>
#include <iterator>

#include "logger.h"
#include "debug.h"
#include "misc.h"
#include "minilibevent.h"
#include "server.h"
#include "Thread.h"
#include "ServerWorkerThread.h"
#include "ServerNetworkThread.h"
#include "ServerModelMetadata.h"
#include "ServerClientConnection.h"


extern unsigned int n_worker_threads;
extern ServerWorkerThread *worker_threads;
extern bool strongly_bind_flag;
extern ServerModelMetadata *model_metadata;
extern unsigned int n_model_metadata;


void shutdown_connection(ServerClientConnection *conn);

ServerNetworkThread::ServerNetworkThread() : Thread()
{
	CRIT_ASSERT(pthread_mutex_init(&evq_mtx, NULL) == 0);
	CRIT_ASSERT(pthread_cond_init(&evq_cv, NULL) == 0);
	CRIT_ASSERT(ev_valid_queue(evq = ev_create_queue()));
	rr_worker_thread = 0;
}

ServerNetworkThread::~ServerNetworkThread()
{
	pthread_cond_destroy(&evq_cv);
	pthread_mutex_destroy(&evq_mtx);
}

#ifndef MAX_NETWORK_EVENTS
#define MAX_NETWORK_EVENTS 40
#endif

/**
 * Network thread - gathers IO from connections and dispatches it to worker
 * threads, in a bursty way: tries to gather as much as MAX_NETWORK_EVENTS
 * events if possible while holding locks to avoid IPC overhead.
 */
void
ServerNetworkThread::run()
{
	ServerWorkerThread *wthr;
	ev_list evs;
	ev_result evr;
	struct timespec ts_wait;
	ServerClientConnection *conn;

	mc_log(LOG_INFO, "Network thread %p starting", this);

	evs = ev_allocate_list(MAX_NETWORK_EVENTS);
	if (evs == NULL)
		mc_log(LOG_CRIT, "Cannot allocate event list in network thread %p", this);

	ts_wait.tv_sec = 0;
	ts_wait.tv_nsec = 100000000; /* 0.1s, 100 ms */
	DEBUG_STATUS_POINT(st, '0');

	while (!stop_requested) {
		DEBUG_STATUS_POINT(st, 'e');
		evr = ev_get_new_events(evq, evs, MAX_NETWORK_EVENTS, &ts_wait);
		/* Process new events in bulk, adding them to worker threads (if applicable) */
		DEBUG_STATUS_POINT(st, 'E');
		for (int i = 0; i < evr; i++) {
			conn = (ServerClientConnection*) evs[i].udata;
			assert(conn != NULL);
			assert(conn->type == CONN_TYPE_CLIENT_UNIX || conn->type == CONN_TYPE_CLIENT_TCP);

			if (evs[i].filter == EV_READ) {
				/* Handle incoming data */
				DEBUG_STATUS_POINT(st, 'r');
				if (conn->process_read_event(this, &model_metadata[index]) == EVPROC_ERROR) {
					ev_remove_fd(evq, conn->socket, FALSE);
					shutdown_connection(conn);
					continue;
				}
			} else if (evs[i].filter == EV_WRITE && conn->outputDataReady()) {
				/* Handle outgoing data */
				DEBUG_STATUS_POINT(st, 'w');
				if (conn->process_write_event() == EVPROC_ERROR) {
					ev_remove_fd(evq, conn->socket, FALSE);
					shutdown_connection(conn);
					continue;
				}
			}
		}

		if (n_worker_threads > 0) {
			wthr = &worker_threads[rr_worker_thread];
			if (wthr->getQueueLocker() == this) {
				wthr->setQueueLocker(NULL);
				wthr->queueUnlock();
				wthr->signalQueueReady();
				if (!strongly_bind_flag) /* If not "strongly bound", rotate to another thread */
					rr_worker_thread = (rr_worker_thread + 1) % n_worker_threads;
			}
		}
	}
	mc_log(LOG_INFO, "Network thread %p exiting", this);
	free(evs);
	DEBUG_STATUS_POINT(st, 'X');
}
